<?php
abstract class Model implements \ArrayAccess {
	/**
	 * @var Criteria
	 */
	private $_criteria;
	
	public $_database = '';
	public $_table = '';
	public $_config;
	protected $_last_query = "";
	/**
	 * @var Database
	 */
	protected $_db;
	
	protected $_primary = "id";
	
	const ORIGEN_VALUE = "__MODEL_ORIGEN_VALUE_!@#$%^&*__I_BELIEVE_YOU_WOULDNOT_USE_THIS";
	
	public function __construct() {
		$this->_table = $this->getTableName();
		if (!isset(AtomCode::$config['database'])) {
			AtomCode::addConfig("database");
		}
		
		$this->_config = & AtomCode::$config['database'][$this->getDatabaseGroup()];

		$this->_db = & Database::get($this->_database);

		$this->reset();
	}

    public function rules() {
	    return [];
    }
	
	public function getCriteria() {
		return clone $this->_criteria;
	}
	
	public function setCriteria($criteria) {
		$this->_criteria = clone $criteria;
	}

	public function offsetSet($offset, $value) {
		if (is_null($offset)) {
			$this->$offset = $value;
		} else {
			$this->$offset = $value;
		}
	}
	public function offsetExists($offset) {
		return isset($this->$offset);
	}
	public function offsetUnset($offset) {
		unset($this->$offset);
	}
	public function offsetGet($offset) {
		return isset($this->$offset) ? $this->$offset : null;
	}
	
	public function __set($name, $value) {
		if ($name{0} != "_") {
			return ;
		}
		
		$this->{$name} = $value;
	}

    public function getDatabaseGroup() {
        if ($this->_database) return $this->_database;

        $class_name = get_called_class();
        $ps = explode("\\", $class_name);
        if (count($ps) == 3) {
            $this->_database = $ps[1];
        } else {
            $this->_database = 'default';
        }

        return $this->_database;
	}
	
	public function getTableName(){
		$class = get_class($this);
		$ps = explode('\\', $class);
		$class = end($ps);
		$name = substr($class, 0, -5);
		$name = preg_replace("/([A-Z])/", "_$1", $name);
		return strtolower(substr($name, 1));
	}
	
	public function getUsingTable() {
		return $this->_config['dbprefix'] . $this->_table;
	}
	
	public function select($columns = '*') {
		$this->_criteria->select = $columns;
	}
	
	public function from($table) {
		$this->_criteria->from = $table;
	}
	
	public function alias($name) {
		$this->_criteria->from = $this->getUsingTable() . ' AS ' . $name;
	}
	
	public function where($where, $binding = array()) {
		if (is_array($where)) {
			foreach ($where as $col => $val) {
				$this->where("$col = :$col");
			}
			
			$this->bind($where);
		} else {
			if (!is_array($binding) && !is_null($binding) && preg_match("/^[\\w\\.`]+$/", $where)) {
				$param = "param_" . ($this->_criteria->counter ++);
				$binding2[$param] = $binding;
				$binding = $binding2;
				
				$where .= " = :" . $param;
			}
			
			$this->_criteria->where[] = $where;
			if ($binding) {
				$this->bind($binding);
			}
		}
	}
	
	/**
	 * 对于分页查询后仍然需要查询总数的情况，需要首先调用本方法
	 */
	public function prepareFoundRows() {
		$this->_criteria->found_rows = true;
	}
	
	public function orderBy($order) {
		$this->_criteria->order = $order;
	}
	
	public function having($having, $binding = array()) {
		$this->bind($binding);
		$this->_criteria->having[] = $having;
	}
	
	public function groupBy($group) {
		$this->_criteria->groupBy = $group;
	}
	
	public function limit($limit, $offset = null) {
		$this->_criteria->limit = ($offset ? $offset . ',' : '') . $limit;
	}
	
	public function data($key, $val = false) {
		if (is_array($key)) {
			$this->inflat($key);
		} else {
			if ($key{0} != '_' && property_exists($this, $key)) {
				$this->{$key} = $val;
			}
		}
	}
	
	public function inflat($values) {
		foreach ($values as $k => $v) {
			if ($k{0} != '_' && property_exists($this, $k)) {
				$this->{$k} = $v;
			}
		}
	}
	
	public function find($binding = array()) {
		$this->bind($binding);

		$this->_last_query = $this->buildSelectSql();
		$result = $this->_db->queryArray($this->_last_query, $this->_criteria->binding);
		
		$this->reset();
		return $result;
	}

    /**
     * 查询出一个来
     * @param bool $wrap
     * @return static
     */
	public function findOne($wrap = false) {
		$this->limit(1);
		$result = $this->find();
		
		if ($result) {
			if ($wrap) {
				return $this->construct($result[0]);
			} else {
				return $result[0];
			}
		} else {
			return null;
		}
	}
	
	private function construct($v) {
		$c = get_class($this);
		$t = new $c();
		$t->inflat($v);
		
		return $t;
	}
	
	public function insertUpdate($data) {
		$this->_last_query = $this->buildInsertUpdateSql($data);
		$result = $this->_db->query($this->_last_query, $this->_criteria->binding);
		
		$this->reset();
		return $result;
	}
	
	public function clean() {
		foreach (get_object_vars($this) as $var => $val) {
			if ($var{0} != "_") {
				$this->{$var} = null;
			}
		}
	}
	
	public function insertSelect($cols = '*') {
		$this->_last_query = $this->buildInsertSelectSql($cols);
	}
	
	/**
	 * @param array $array
	 * @return boolean
	 */
	public function insertBatch($array, $ignore = false) {
		$this->_last_query = $this->buildInsertBatchSql($array, $ignore);
		$result = $this->_db->query($this->_last_query, $this->_criteria->binding);

		$this->reset();
		return $result;
		
	}
	
	public function leftJoin($table, $cond) {
		$this->_criteria->join[] = array("LEFT", $table, $cond);
	}
	
	public function rightJoin($table, $cond) {
		$this->_criteria->join[] = array("RIGHT", $table, $cond);
	}
	
	public function innerJoin($table, $cond) {
		$this->_criteria->join[] = array("INNER", $table, $cond);
	}
	
	public function outerJoin($table, $cond) {
		$this->_criteria->join[] = array("OUTER", $table, $cond);
	}

    /**
     * @param $id
     * @return static
     */
	public function get($id) {
		$this->where($this->_primary . ' = :' . $this->_primary, array($this->_primary => $id));
		$this->limit(1);
		$rows = $this->find();
		
		return $rows ? $this->construct($rows[0]) : null;
	}
	
	public function insert($data = array(), $ignore = false) {
		$this->_last_query = $this->buildInsertSql($data ? $data : $this->value(), $ignore);
		$result = $this->_db->query($this->_last_query, $this->_criteria->binding);

		$this->reset();
		if (property_exists($this, $this->_primary) && !$this->{$this->_primary}) {
			$this->{$this->_primary} = $this->lastInsertId();
		}
		
		return $result;
	}
	
	public function update($data = array()) {
		$this->_last_query = $this->buildUpdateSql($data ? $data : $this->value());
		$result = $this->_db->query($this->_last_query, $this->_criteria->binding);

		$this->reset();
		return $result;
	}
	
	public function delete($binding = array()) {
		$this->bind($binding);
		
		if ($this->{$this->_primary}) {
			$this->where("{$this->_primary}=:primary_key", array('primary_key' => $this->{$this->_primary}));
		}
		
		$this->_last_query = $this->buildDeleteSql();
		$result = $this->_db->query($this->_last_query, $this->_criteria->binding);

		$this->reset();
		return $result;
	}
	
	public function save() {
		$data = $this->value();
		
		if ($this->{$this->_primary}) {
			$this->where("`{$this->_primary}`=:primary_key", array('primary_key' => $this->{$this->_primary}));
		}
		
		$result = !!$this->insertUpdate($data);
		$this->reset();
		if (property_exists($this, $this->_primary) && !$this->{$this->_primary}) {
			$this->{$this->_primary} = $this->lastInsertId();
		}
		
		return $result;
	}
	
	public function lock($cond = '') {
		if (!$cond) {
			$cond = $this->getUsingTable() . " READ";
		}
		$this->_last_query = $this->buildLockSql($cond);
		$result = $this->_db->query($this->_last_query);
		return $result;
	}
	
	public function unlock() {
		$this->_last_query = $this->buildUnlockSql();
		$result = $this->_db->query($this->_last_query);
		return $result;
	}
	
	public function bind($key, $value = null) {
		if (!$key) return ;
		
		if (is_array($key)) {
			foreach ($key as $k => $v) {
				if (is_array($v)) {
					$v = $this->_getValueArray($v);
				}
				
				$this->_criteria->binding[$k] = $v;
			}
		} else {
			if (is_array($value)) {
				$value = $this->_getValueArray($value);
			}
			
			$this->_criteria->binding[$key] = $value;
		}
	}
	
	/**
	 * 
	 * @param array $arr
	 */
	private function _getValueArray($arr) {
		return implode(", ", $this->quote($arr));
	}
	
	public function query($sql, $binding = array()) {
		$this->bind($binding);
		$this->_last_query = $sql;
		$result = $this->_db->query($this->_last_query, $this->_criteria->binding);

		$this->reset();
		return $result;
	}
	
	public function queryArray($sql, $binding = array()) {
		$this->bind($binding);
		$this->_last_query = $sql;
		$result = $this->_db->queryArray($this->_last_query, $this->_criteria->binding);

		$this->reset();
		return $result;
	}
	
	public function buildSelectSql() {
		return $this->partSelectSql() . $this->partFromSql() . $this->partJoinSql()
		 . $this->partWhereSql() . $this->partGroupSql() . $this->partHavingSql()
		 . $this->partOrderSql() . $this->partLimitSql() . $this->partForUpdateSql();
	}
	
	public function buildInsertSelectSql($cols = '*') {
		if ($cols == '*') {
			$cols_sql = "";
		} else {
			$cols_sql = "($cols)";
		}
	}
	
	public function buildInsertUpdateSql($data) {
		if (isset($data[$this->_primary])) {
			unset($data[$this->_primary]);
		}
		return $this->buildInsertSql($this->value()) . ' ON DUPLICATE KEY UPDATE ' . $this->partUpdateSql($data);
	}
	
	public function buildInsertBatchUpdateSql($array, $data) {
		return $this->buildInsertBatchSql($array) . ' ON DUPLICATE KEY UPDATE ' . $this->partUpdateSql($data);
	}
	
	public function buildInsertSql($data, $ignore = false) {
		foreach ($data as $k => $v) {
			if (is_null($v)) {
				unset($data[$k]);
			}
		}
		$cols = array_keys($data);
		$cols = $this->quoteKey($cols);
		
		return 'INSERT' . ($ignore ? ' IGNORE' : '') . ' INTO ' . $this->getUsingTable() . ' (' . implode(', ', $cols) . ') VALUES (' . implode(', ', $this->quote($data)) . ')';
	}
	
	public function buildInsertBatchSql($array, $ignore = false) {
		if (!count($array)) {
			throw new Exception("nothing to be inserted!");
		}

		$values = array();
		// build data array according to the first element.
		// so, we detect the first element's values
		$cols = array();
		$col_in_key = array();
		foreach ($array[0] as $k => $v) {
			if (is_null($v)) {
				unset($array[0][$k]);
			}
			
			$cols[] = $k;
		}
		$this->quoteKey($cols);
		
		// now we apply a filter for strip uneffective keys in the array 
		foreach ($array as $data) {
			$data2 = array();
			foreach ($cols as $k) {
				$data2[$k] = $data[$k];
			}
			
			$values[] = '(' . implode(', ', $this->quote($data2)) . ')';
		}
		
		return 'INSERT' . ($ignore ? ' IGNORE' : '') . ' INTO ' . $this->getUsingTable() . ' (' . implode(', ', $cols) . ') VALUES ' . implode(", ", $values);
		
	}
	
	public function buildUpdateSql($data) {
		$data = $data ? $data : $this->value();
		
		return 'UPDATE ' . $this->partFromSql(false) . $this->partJoinSql() . ' SET ' . $this->partUpdateSql($data) . $this->partWhereSql() . $this->partOrderSql() . $this->partLimitSql();
	}
	
	public function buildDeleteSql() {
		return 'DELETE ' . $this->partFromSql() . $this->partWhereSql() . $this->partOrderSql() . $this->partLimitSql();
	}
	
	public function buildLockSql($read) {
		$sql = 'LOCK TABLE ' . $read;
		return $sql;
	}
	
	public function buildUnlockSql() {
		return 'UNLOCK TABLES';
	}
	
	private function partSelectSql() {
		return 'SELECT ' . ($this->_criteria->found_rows ? 'SQL_CALC_FOUND_ROWS ' : '') . ($this->_criteria->select ? $this->_criteria->select : implode(', ', $this->quoteKey($this->getTableColumns())));
	}
	
	private function partFromSql($add_from = true) {
		return ($add_from ? ' FROM ' : ' ') . ($this->_criteria->from ? $this->_criteria->from : $this->getUsingTable());
	}
	
	private function partJoinSql() {
		$join_str = array();
		foreach ($this->_criteria->join as $row) {
			if (count($row) == 3) {
				$join_str[] = $row[0] . " JOIN " . $row[1] . ($row[2] ? " ON " . $row[2] : "");
			}
		}
		
		return ' ' . implode(" ", $join_str);
	}
	
	private function partWhereSql() {
		if ($this->_criteria->where) {
			if (count($this->_criteria->where) > 1) {
				return ' WHERE (' . implode(') AND (', $this->_criteria->where) . ')';
			} else {
				return ' WHERE ' . implode(' AND ', $this->_criteria->where);
			}
			
		} else {
			return '';
		}
	}
	
	private function partGroupSql() {
		if ($this->_criteria->groupBy) {
			return ' GROUP BY ' . $this->_criteria->groupBy;
		} else {
			return '';
		}
	}
	
	private function partHavingSql() {
		if ($this->_criteria->having) {
			if (count($this->_criteria->having) > 1) {
				return ' HAVING (' . implode(') AND (', $this->_criteria->having) . ')';
			} else {
				return ' HAVING ' . implode(' AND ', $this->_criteria->having);
			}
		} else {
			return '';
		}
	}
	
	private function partOrderSql() {
		if ($this->_criteria->order) {
			return ' ORDER BY ' . $this->_criteria->order;
		} else {
			return '';
		}
	}
	
	private function partLimitSql() {
		if ($this->_criteria->limit) {
			return ' LIMIT ' . $this->_criteria->limit;
		} else {
			return '';
		}
	}

    private function partForUpdateSql() {
	    if ($this->_criteria->for_update) {
	        return ' For UPDATE';
        } else {
	        return '';
        }
	}
	
	private function partUpdateSql($data) {
		$items = array();
		foreach ($data as $k => $v) {
			if (is_null($v)) continue;
			
			if ($v === self::ORIGEN_VALUE) {
				$items[] = $k;
			} else {
				$k = $this->quoteKey($k);
				
				$items[] = $k . '=' . ($v == '?' ? $v : $this->quote($v));
			}
		}
		
		return implode(', ', $items);
	}
	
	public function quote($data) {
		if (is_array($data)) {
			foreach ($data as &$d) {
				$d = addslashes($d);
				$d = "'$d'";
			}
		} else {
			$data = addslashes($data);
			$data = "'$data'";
		}
		
		return $data;
	}
	
	public function quoteKey($keys) {
		if (!is_array($keys)) {
			if (preg_match("/^\w+$/", $keys) && !is_numeric($keys{0})) {
				return '`' . $keys . '`';
			} else {
				return $keys;
			}
		}

		if (array_key_exists(0, $keys)) {
			
			foreach ($keys as &$key) {
				$key = $this->quoteKey($key);
			}
			
			return $keys;
		}
		
		$new = array();
		foreach ($keys as $key => $value) {
			$key = $this->quoteKey($key);
			
			$new[$key] = $value;
		}
		
		return $new;
	}
	
	public function value() {
		$c = array();
		$ps = get_object_vars($this);
		foreach ($ps as $n => $_) {
			if ($n{0} != '_') {
				$c[$n] = $_;
			}
		}

		return $c;
	}
	
	protected function getTableColumns() {
		return array_keys($this->value());
	}
	
	public function getMessages() {
		return $this->_db->getErrors();
	}
	
	public function getLastQuery() {
		return $this->_last_query;
	}
	
	public function affectRows() {
		return $this->_db->affectRows();
	}
	
	public function lastInsertId() {
		return $this->_db->lastInsertId();
	}
	
	public function reset() {
		$this->_criteria = new Criteria();
	}
	
	/**
	 * 
	 * @return 查询 limit 语句中去掉 limit 语句的总条数
	 */
	public function foundRows() {
		$result = $this->_db->query("SELECT FOUND_ROWS() count");
		$all = $result->fetchColumn();
		
		return $all;
	}
	
	/**
	 * @link Model::foundRows()
	 */
	public function getFoundRows() {
		return $this->foundRows();
	}
	
	/**
	 * 启动一次事务
	 * 
	 * 如果之前已经启动了一次事务，则不会有任何反应。启动一次事务的本质是关闭自动提交事务模式。
	 * 
	 * 如果启动了事务但中途中断执行而没有调用提交事务，则会自动回滚。
	 * 
	 * 请不要同时使用 mysql_query("begin transaction") 或 $model->query("begin transaction"); 或其他方式启动事务。否则将在每个链接单独启用事务。
	 * 
	 * @param int 错误类型
	 * @return bool Returns true on success or false on failure.  
	 * @link http://www.php.net/manual/zh/pdo.begintransaction.php
	 * @link http://php.net/manual/zh/pdo.constants.php#PDO::ERRMODE_SILENT
	 */
	public function beginTransaction($err_mode = PDO::ERRMODE_WARNING) {
		return $this->_db->beginTransaction($err_mode);
	}
	
	/**
	 * 提交事务，务必在启用事务后提交，否则没有任何作用
	 */
	public function commit() {
		return $this->_db->commit();
	}
	
	/**
	 * 回滚事务
	 */
	public function rollback() {
		return $this->_db->rollback();
	}

    public function forUpate($set = true)
    {
        $this->_criteria->for_update = $set ? true : false;
	}
}